#include "Hashtable.h"
#include <iostream.h>




class HashCons
{
public:
	ulong Key;
	Object* pEntry;
	HashCons* pNext;
	HashCons( ulong k, Object* v )
	{ Key = k, pEntry = v, pNext = null; }
};


Object* Hashtable::start(ulong& x, ulong& y)
{
	x = 0, y = 0;
	advance( x, y );
	return item( x, y );
}


Object* Hashtable::item( ulong x, ulong y )
{
	if ( x >= tableSize ) return null;

	HashCons* hc = table[ x ];
	for ( int i = 0; i < y; i ++ ) hc = hc->pNext;
	
	return hc->pEntry;
}


void Hashtable::remove( ulong x, ulong y )
{
	if ( x >= tableSize ) return;

	HashCons* hc = table[ x ];
	for ( int i = 0; i < y; i ++ ) hc = hc->pNext;
	
	remove( hc->Key );
}


void Hashtable::advance( ulong& x, ulong& y )
{
	while( table[ x ] == null && x < tableSize ) x++;
	if ( x >= tableSize ) return;
	
	HashCons* hc = table[ x ];
	
	for ( int i = 0; i < y; i ++ ) hc = hc->pNext;

	if ( hc->pNext == null )
	{
		for ( x++; table[ x ] == null && x < tableSize; x++ );
		y = 0;
	}
	else y++;
}


ulong Hashtable::getIndex( ulong Key )
{
	return ( ( ulong ) Key ) % tableSize;
}


bool Hashtable::contains( ulong Key )
{
	ulong i = getIndex( Key );
	HashCons* e = table[ i ];

	while ( e != null && e->Key != Key )
	{
		e = e->pNext;
	}

	return e != null;
}


Object* Hashtable::item( ulong Key )
{
	ulong i = getIndex( Key );
	HashCons* e = table[ i ];

	while ( e != null && e->Key != Key )
	{
		e = e->pNext;
	}

	if ( e != null ) return e->pEntry;
	else return null;
}


bool Hashtable::isEmpty()
{
	return size() == 0;
}


ulong Hashtable::size()
{
	return numberOfEntries;
}


bool Hashtable::insert( ulong Key, Object* Object )
{
	ulong i = getIndex( Key );

	if ( table[ i ] == null )
	{
		table[ i ] = new HashCons( Key, Object );
		if ( table[ i ] == null ) return false;
		numberOfLists++;
		numberOfEntries++;
	}
	else
	{
		HashCons * e = table[ i ];

		while ( e->Key != Key && e->pNext != null )
		{
			e = e->pNext;
		}

		if( e->Key == Key )
		{
			cerr << "Hashtable::insert(): Key " << Key << " existiert schon!\n";
			return false;
		}

		e->pNext = new HashCons( Key, Object );
		if ( e->pNext == null ) return false;
		numberOfEntries++;
	}

	rehashTest();

	return true;
}


bool Hashtable::remove( ulong Key )
{
	ulong i = getIndex( Key );
	HashCons* e = table[ i ];
	HashCons* v = e;

	while ( e != null && e->Key != Key )	// Schluessel suchen...
	{
		v = e;
		e = e->pNext;
	}

	if ( e == null )						// Schluessel gefunden?
	{
		return false;
	}

	if ( v != e )							// Schluessel unter mehreren gefunden
	{
		v->pNext = e->pNext;
	}
	else									// Schluessel als einziger in der Liste
	{
		table[ i ] = e->pNext;				// Liste austragen
	}

	delete e;								// und Eintrag loeschen
	numberOfEntries--;
	if ( table[ i ] == null ) numberOfLists--;

	rehashTest();

	return true;							// wie sollte es anders sein?
}


ulong Hashtable::findTableSize( ulong size )
{
	static const ulong hashSize[] =
	{
		23, 53, 101, 211, 503,
		1009, 2003, 5003, 10007, 20011,
		50021L, 100003L, 0
	};

	ulong i = 0;
	for ( ; hashSize[ i ] != 0; i++ )
	{
		if ( hashSize[ i ] > size ) break;
	}

	if ( hashSize[ i ] > size ) return hashSize[ i ];
	else return size % 4 + 13;				// Das ist nur ein Notbehelf
}


Hashtable::Hashtable( ulong Size )
{
	numberOfEntries = 0;
	numberOfLists   = 0;
	tableSize	  	= findTableSize( Size );

	table = new ( HashCons* [ tableSize ] );

	if ( table == null )
	{
    	cerr << "Hashtable::Hashtable(): Kein Speicher fuer Tabelle vorhanden\n";
		tableSize = 0;
		return;
	}

	for ( ulong i = 0; i < tableSize; i++) table[ i ] = null;
}


Hashtable::~Hashtable()
{
	for ( ulong i = 0; i < tableSize; i++ )
	{
		for ( HashCons* e = table[ i ]; e != null; )
		{
			HashCons *h = e;
			e = e->pNext;
			delete h;
		}
	}

	delete[] table;
}


void Hashtable::rehashInsert( HashCons* e )
{
	ulong i = getIndex( e->Key );

	if ( table[ i ] == null )
	{
		table[ i ] = e;
		numberOfLists++;
	}
	else
	{
		HashCons* h = table[i];
		for ( ; h->pNext != null; h = h->pNext );
		h->pNext = e;
		e->pNext = null;
	}
}


void Hashtable::rehashTest()
{
	if ( ( numberOfEntries - numberOfLists ) > numberOfEntries )
	{
		rehash();
	}
}


void Hashtable::rehash()
{
	ulong oldSize = tableSize;
	HashCons** pOldTable = table;

	ulong newSize = findTableSize( numberOfEntries );
	if ( newSize <= tableSize )
	{
		newSize = findTableSize( tableSize + 1 );
	}

	table = new ( HashCons*[ newSize ] );
	if ( table == null )
	{
		table = pOldTable;
		return;
	}

	ulong i = 0;
	for ( ; i < newSize; i ++ ) table[ i ] = null;
	numberOfLists = 0;
	tableSize = newSize;

	for( i = 0; i < oldSize; i++)
	{
		for ( HashCons *e = pOldTable[ i ]; e != null; )
		{
			HashCons* n = e->pNext;
			rehashInsert( e );
			e = n;
		}
	}

	delete[] pOldTable;
}

